#include "CyanLight.hpp"

// C++ system level
#include <cstdlib>  // TODO: was this for std::min() ?

// ESP32 specific
#include "esp_log.h"


uint8_t MAX_CHANNELS = 8;

const char* TAG = "CyanLight";

const uint8_t channelGPIOs[] = { 27, 16, 17, 18, 19, 21, 22, 23 };

namespace CyanLight {

  CyanLightControl::CyanLightControl(uint8_t channelsConfigured, uint32_t baseChannel) : SiliconTorch::FxCyanF::FxCyanF(baseChannel) {

    uint8_t channels = std::min(channelsConfigured, MAX_CHANNELS);

    for (uint8_t i = 0; i < channels; i++) this->addChannel();
  }


  bool CyanLightControl::addChannel() {

    auto ch = this->getChannelCount();

    if (ch < MAX_CHANNELS) {
      return SiliconTorch::FxCyanF::FxCyanF::addChannel((uint8_t)channelGPIOs[ch]);
    } else {
      ESP_LOGE(TAG, "Cannot create channel#[ %i ]! CLC hardware limit is #[ %i ]", ch, MAX_CHANNELS);
      return false;
    }
  }
}
