#include "FxCyanF.hpp"

// C++ system level
#include <string>
#include <cstdio>         // snprintf
// #include <cstdlib>
// #include <algorithm>
// #include <functional>

// ESP32 specific
#include "nvs.h"
#include "nvs_flash.h"
#include "esp_log.h"

// misc
// #include <nlohmann/json.hpp>

// qthing stuff
// #include <qthing>
// #include <qthing/mqtt_common.hpp>

// project specific
// #include "FxVSync.hpp"
// #include "CyanBus.hpp"
// #include "Metrics.hpp"



namespace SiliconTorch {

  namespace FxCyanF {



    bool configureFromNVS(FxCyanF& target) {

      if (target.getChannelCount() > 0) {
        ESP_LOGW(TAG, "Target mustn't have any channels configured");
        return false;
      }


      nvs_flash_init();

      nvs_handle_t nvs;
      nvs_open(nvsNameSpace, NVS_READONLY, &nvs);


      uint32_t frq = 0;
      nvs_get_u32(nvs, "frequency", &frq);

      uint8_t res = 0;
      nvs_get_u8(nvs, "resolution", &res);


      if (frq == 0 || res == 0) {
        ESP_LOGW(TAG, "No config for CyanLight found");  // TODO: Log-Level Debug
        return false;
      }


      ESP_LOGW(TAG, "Found config for FxCyanF{ frequency[ %d Hz ]  resolution[ %d Bit ] } in NVS", frq, res);  // TODO: Log-Level Debug

      for (uint8_t ch = 0; ch < MAX_CHANNELS; ch++) {

        int8_t gpio = -1;
        
        char channel[32];
        snprintf(channel, 32, "ch%d_gpio", ch);

        nvs_get_i8(nvs, channel, &gpio);

        if (gpio >= 0) {

          target.addChannel((uint8_t)gpio);

          ESP_LOGW(TAG, "channel[ %d ]  gpio[ %d ]", ch, gpio);  // TODO: Log-Level Info

        } else {
          break;
        }
      }

      return true;
    }


  }
}
