#include "NVSExplorer.hpp"

// C++ system level
#include <set>
#include <vector>
#include <cstring>     // memset, memcpy, strncmp
// #include <cstdlib>
#include <cinttypes>
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "nvs_flash.h"
#include "esp_err.h"
// #include "driver/uart.h"

// project specific
#include <SpiderLib/Util.hpp>

// qthing stuff
#include <qthing>
#include <qthing/mqtt_common.hpp>




namespace SiliconTorch {

  namespace NVSExplorer {


    const char* TAG = "NVSExplorer";


    NVSExplorer::NVSExplorer() {

      ESP_LOGI(TAG, "Starting NVSExplorer[📡]");

      nvs_flash_init();

      uint8_t l = 0;
      uint8_t u8 = 88;
      uint64_t u64 = 1337;

      const char* err;

      /* err = esp_err_to_name( nvs_open("__test__", NVS_READWRITE, &nvs) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]", l++, err);

      err = esp_err_to_name( nvs_get_u8(nvs, "u8", &u8) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  get_u8[ %d ]", l++, err, u8);

      u8 = 88;
      err = esp_err_to_name( nvs_set_u8(nvs, "u8", u8) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  set_u8", l++, err);

      err = esp_err_to_name( nvs_get_u8(nvs, "u8", &u8) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  get_u8[ %d ]", l++, err, u8);

      ESP_ERROR_CHECK( nvs_commit(nvs) );

      err = esp_err_to_name( nvs_get_u64(nvs, "u8", (uint64_t*)(&u8)) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  get_u64[ %d ]", l++, err, u8);

      u8 = 88;
      err = esp_err_to_name( nvs_set_u64(nvs, "u8", (uint64_t)u64) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  set_u64", l++, err);

      err = esp_err_to_name( nvs_get_u64(nvs, "u8", (uint64_t*)(&u64)) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  get_u64[ %lld ]", l++, err, u64);

      ESP_ERROR_CHECK( nvs_commit(nvs) ); */

      //const char* key = "__key__";
      const char* key = "u8";

      char* out_value = new char[128];
      size_t size;
      sprintf(out_value, "__empty_value__");

      err = esp_err_to_name( nvs_open("__test__", NVS_READWRITE, &nvs) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  open()", l++, err);

      //err = esp_err_to_name( nvs_erase_key(nvs, key) );
      //ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  erase()", l++, err);

      //err = esp_err_to_name( nvs_commit(nvs) );
      //ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  commit()", l++, err);

      err = esp_err_to_name( nvs_get_str(nvs, key, out_value, &size) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  get_str[ %s ]", l++, err, out_value);

      err = esp_err_to_name( nvs_set_str(nvs, key, out_value) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  set_str[ %s ]", l++, err, out_value);

      err = esp_err_to_name( nvs_commit(nvs) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  commit()", l++, err);

      sprintf(out_value, "==other-empty==");
      err = esp_err_to_name( nvs_get_str(nvs, key, out_value, &size) );
      ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]  get_str[ %s ]", l++, err, out_value);


      // err = esp_err_to_name(  );
      // ESP_LOGW(TAG, "Line[ %d ]  Error[ %s ]", l++, err);



      uint16_t deviceTopicLEN = deviceTopic("").length();

      auto wrapMessageHandler = [deviceTopicLEN](std::function<void(std::vector<std::string>&, qthing::multipart_message_t)> handler) {
        return [handler, deviceTopicLEN](qthing::multipart_message_t msg) {
          if (msg.length == msg.total_length && msg.offset == 0) {

            std::string localTopic(msg.topic.c_str() + deviceTopicLEN);
            std::vector<std::string> topicParts = SpiderLib::Util::split(localTopic, '/');

            handler(topicParts, msg);
          } else {
            ESP_LOGW(TAG, "Multipart messages are not supported atm.  topic[ %s ]", msg.topic.c_str());
          }
        };
      };


      auto lsKeys = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        
        std::string delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = std::string(buffer);
        }

        auto keys = listKeys(topic[1]);
        std::string msgOut = SpiderLib::Util::join(keys, delimiter);

        // TODO: return topic…?
        qthing::publish_message(deviceTopic(topic[1]), msgOut);

        if (buffer != NULL) delete buffer;
      };

      // TODO: return topic…?
      qthing::add_binary_message_callback(deviceTopic("ls/+"), wrapMessageHandler(lsKeys));


      auto lsNS = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        
        std::string delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = std::string(buffer);
        }

        auto nameSpaces = listNamespaces();
        std::string msgOut = SpiderLib::Util::join(nameSpaces, delimiter);

        // TODO: return topic…?
        qthing::publish_message(deviceTopic(topic[0]), msgOut);

        if (buffer != NULL) delete buffer;
      };

      // TODO: return topic…?
      qthing::add_binary_message_callback(deviceTopic("ls"), wrapMessageHandler(lsNS));



      auto rmKey = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        this->rmKey(topic[0], topic[1]);
      };

      qthing::add_binary_message_callback(deviceTopic("+/+/rm"), wrapMessageHandler(rmKey));


    }


    void NVSExplorer::BullShit() {

      ESP_LOGW(TAG, "Scanning 👀");

      mutex.run([&]() {
        nvs_iterator_t it = nvs_entry_find("nvs", NULL, NVS_TYPE_ANY);

        while (it != NULL) {
          nvs_entry_info_t info;
          nvs_entry_info(it, &info);
          it = nvs_entry_next(it);

          ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);
        }
      });
    }


    void NVSExplorer::rmKey(const std::string& nameSpace, const std::string& key) {
      mutex.run([&]() {
        nvs_handle_t _nvs;
        ESP_LOGW(TAG, "rmKey() -> %s / %s", nameSpace.c_str(), key.c_str());
        ESP_LOGW(TAG, "open() -> %s", esp_err_to_name( nvs_open(nameSpace.c_str(), NVS_READWRITE, &_nvs) ));
        ESP_LOGW(TAG, "erase() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
        ESP_LOGW(TAG, "commit() -> %s", esp_err_to_name( nvs_commit(_nvs) ));
        ESP_LOGW(TAG, "erase²() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
      });
    }

    std::set<std::string> NVSExplorer::listKeys(const std::string& nameSpace) {

      std::set<std::string> keys;

      mutex.run([&]() {
        nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

        while (it != NULL) {
          nvs_entry_info_t info;
          nvs_entry_info(it, &info);
          it = nvs_entry_next(it);

          ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);  // TODO: remove after dev!

          keys.insert(info.key);
        }
      });

      return keys;
    }

    std::set<std::string> NVSExplorer::listNamespaces() {

      std::set<std::string> namespaces;


      mutex.run([&]() {
        nvs_iterator_t it = nvs_entry_find("nvs", NULL, NVS_TYPE_ANY);

        while (it != NULL) {
          nvs_entry_info_t info;
          nvs_entry_info(it, &info);
          it = nvs_entry_next(it);

          ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);  // TODO: remove after dev!

          namespaces.insert(info.namespace_name);
        }
      });

      return namespaces;
    }

    void NVSExplorer::switchNS(const std::string& newNS) {
      mutex.run([&]() {
        // if (currentNS != NULL && currentNS.compare(newNS) != 0) {

          // currentNS = newNS;



          // esp_err_t err = nvs_open("__test__", NVS_READWRITE, &nvs);

        // }
      });
    }

    std::string NVSExplorer::deviceTopic(const std::string& suffix) {
      return std::string(DEVICE_NAMESPACE + "NVS/") + suffix;
    }

  }
}
