#pragma once

// C++ system level
#include <set>
#include <cstring>     // memset, strncmp
// #include <cstdlib>     // TODO: is this for memcpy?
#include <cinttypes>
// #include <functional>

// ESP32 specific
// #include "esp_err.h"
#include "nvs_flash.h"
#include "esp_log.h"
// #include "driver/gpio.h"
// #include "driver/uart.h"

// project specific
#include "SpiderLib/ManagedMutex.hpp"

// qthing stuff
// #include <qthing>


namespace SiliconTorch {

  namespace NVSExplorer {


    extern const char* TAG;


    class NVSExplorer {

      public:

        void BullShit();


        std::set<std::string> listNamespaces();
        std::set<std::string> listKeys(const std::string& nameSpace);

        void rmKey(const std::string& nameSpace, const std::string& key);

        // Gets the key regardless of its type in flash
        int64_t  getSignedInt(const std::string& nameSpace, const std::string& key);
        uint64_t getUnsignedInt(const std::string& nameSpace, const std::string& key);




        std::string deviceTopic(const std::string& suffix);

        // ⬇⬇ Singleton stuff ⬇⬇

        static NVSExplorer& instance() {
          static NVSExplorer _instance;
          return _instance;
        }

        NVSExplorer(NVSExplorer const&)    = delete;
        void operator=(NVSExplorer const&) = delete;

        // ⬆⬆ Singleton stuff ⬆⬆

      private:

        NVSExplorer();

        nvs_handle_t nvs;

        std::string currentNS;

        SpiderLib::ManagedMutex mutex;

        void switchNS(const std::string& newNS);

    };


  }
}
